#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

# The README.md will be used as the content for the PyPi package details page on the Python Package Index.
with open('README.md', 'r') as readme:
    long_description = readme.read()


setup(
    name='polyswarm-api',
    version='2.0.3.dev2',
    description='Client library to simplify interacting with the PolySwarm consumer API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='PolySwarm Developers',
    author_email='info@polyswarm.io',
    url='https://github.com/polyswarm/polyswarm-api',
    license='MIT',
    python_requires='>=2.7,<4',
    install_requires=[
        'requests~=2.22.0',
        'jsonschema~=3.0.2',
        'ordered-set~=3.1.1',
        'future~=0.18.2',
        'python-dateutil~=2.8.1',
    ],
    extras_require={':python_version < "3.0"': ['futures==3.3.0', 'enum34==1.1.6']},
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ]
)
